/**
 ****************************************************************************************
 *
 * @file app.c
 *
 * @brief Application entry point
 *
 * Copyright (C) RivieraWaves 2009-2015
 *
 *
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @addtogroup APP
 * @{
 ****************************************************************************************
 */

/*
 * INCLUDE FILES
 ****************************************************************************************
 */

#include "rwip_config.h" // SW configuration

#if (BLE_APP_PRESENT)
#include <string.h>
//#include "rwapp_config.h"
#include "app_task.h"  // Application task Definition
#include "app.h"       // Application Definition
#include "gap.h"       // GAP Definition
#include "gapm_task.h" // GAP Manager Task API
#include "gapc_task.h" // GAP Controller Task API

#include "co_bt.h"   // Common BT Definition
#include "co_math.h" // Common Maths Definition
#include "ke_timer.h"
#include "app_dis.h"  // Device Information Service Application Definitions
#include "app_batt.h" // Battery Application Definitions
#if BLE_CUSTOM_SERVER
#include "app_custom.h"  // custom profile Definition
#endif
#include "app_midi.h" // BLE MIDI Definition
#if (NVDS_SUPPORT)
#include "nvds.h" // NVDS Definitions
#endif
#include "rf.h"
#include "uart.h"
#include "uart2.h"
#include "adc.h"
#include "gpio.h"
#include "wdt.h"
#include "rtc.h"
#include "user_config.h"
#include "flash.h"
#include "ke_mem.h"

/*
 * DEFINES
 ****************************************************************************************
 */
struct SRAM_USR_FLASH_DATA_t
{
    uint8_t info;
    uint8_t data;
};

#define TAG_ADV_LEN 1
#define TAG_BAUD_LEN 1
#define TAG_NAME_LEN 20
#define TAG_FFF0_LEN 2
#define TAG_FFF1_LEN 2
#define TAG_FFF2_LEN 2

#define FFF0_LOAD_MASK 0x01
#define FFF1_LOAD_MASK 0x02
#define FFF2_LOAD_MASK 0x04
//#define APP_DEVICE_NAME_LENGTH_MAX      (20)//(18)
char Sram_Device_name[APP_DEVICE_NAME_LENGTH_MAX] = {"WB415-GATT"}; // fix1 2018/12/15 Serial-GATT-->SerialSPP
uint16_t Sram_Adv_interval = 160;
uint16_t Sram_FFF0_Service = 0xFFF0;
uint16_t Sram_FFF1_Char_Service = 0xFFF1;
uint16_t Sram_FFF2_Char_Service = 0xFFF2;

uint16_t adv_service_uuid = 0xF000;
uint16_t adv_char_uuid1 = 0xF001;
uint16_t adv_char_uuid2 = 0xF002;

uint8_t Sram_FFFx_load_from_flash = 0x07; // 0x01 FFF0 do not load from flash,0x02 FFF1 do not load from flash,0x04 FFF2 do not load from flash

uint8_t Sram_Baudrate = 1;
uint8_t scan_response_data[APP_DEVICE_NAME_LENGTH_MAX + 2];
// struct SRAM_USR_FLASH_DATA_t BK3432_used_sram_data_info[128];
struct SRAM_USR_FLASH_DATA_t extMCU_used_sram_data_info[128];

at_rsp_content rsp_result = {0};
bool uart_tp_mode_flag = false;
/*
 * TYPE DEFINITIONS
 ****************************************************************************************
 */

typedef void (*appm_add_svc_func_t)(void);

/*
 * ENUMERATIONS
 ****************************************************************************************
 */

/// List of service to add in the database
enum appm_svc_list
{

    APPM_SVC_UART_TP,

#if BLE_CUSTOM_SERVER
    APPM_SVC_CUSTOM,
#endif
#if BLE_FFF0_SERVER
    APPM_SVC_FFF0,
#endif
    APPM_SVC_DIS,
    APPM_SVC_BATT,
#if BLE_OADS_SERVER
    APPM_SVC_OADS,
#endif
    APPM_SVC_LIST_STOP,
};

/*
 * LOCAL VARIABLES DEFINITIONS
 ****************************************************************************************
 */

/// Application Task Descriptor
static const struct ke_task_desc TASK_DESC_APP = {NULL, &appm_default_handler,
                                                  appm_state, APPM_STATE_MAX, APP_IDX_MAX};

/// List of functions used to create the database
static const appm_add_svc_func_t appm_add_svc_func_list[APPM_SVC_LIST_STOP] =
    {

        (appm_add_svc_func_t)app_midi_add_midis,
#if BLE_CUSTOM_SERVER
        (appm_add_svc_func_t)app_custom_add_customs,
#endif
#if BLE_FFF0_SERVER
        (appm_add_svc_func_t)app_fff0_add_fff0s,
#endif
        (appm_add_svc_func_t)app_dis_add_dis,
        (appm_add_svc_func_t)app_batt_add_bas,
#if BLE_OADS_SERVER
        (appm_add_svc_func_t)app_oad_add_oads,
#endif
};

/*
 * GLOBAL VARIABLE DEFINITIONS
 ****************************************************************************************
 */

/// Application Environment Structure
struct app_env_tag app_env;

/*
 * FUNCTION DEFINITIONS
 ****************************************************************************************
 */
uint8_t appm_set_dev_name(char *name, uint8_t len);
void NVR_extMCU_USED_FLASH_AREA_info_get(void);
void NVR_BK3432_USED_FLASH_AREA_info_get(void);
void load_parameter_from_BK3432_USED_FLASH_AREA(void);
void WFLASH_process(uint8_t address, uint8_t data, uint16_t area);
uint8_t RFLASH_process(uint8_t address, uint16_t area_addr);
uint8_t search_TAG_empty_page_no(uint8_t TAG_value);
void Get_page_info_data(uint16_t area, uint8_t page_no, uint8_t *data, uint8_t *info);
void update_page_info_data(uint16_t area, uint8_t page_no, uint8_t info, uint8_t data);
uint8_t flash_unit_info_transf_data_idx(uint8_t info);
uint8_t Get_page_info_data_in_FLASH(uint16_t area, uint8_t page_no);
void app_fff1_send_lvl(uint8_t *buf, uint8_t len);
void app_midi_send_payload(uint8_t *buf, uint8_t len);
void appm_init()
{
    // Reset the application manager environment
    memset(&app_env, 0, sizeof(app_env));

    // Create APP task
    ke_task_create(TASK_APP, &TASK_DESC_APP);

    // Initialize Task state
    ke_state_set(TASK_APP, APPM_INIT);

#if (NVDS_SUPPORT)
    uint8_t key_len = KEY_LEN;
    if (nvds_get(NVDS_TAG_LOC_IRK, &key_len, app_env.loc_irk) != NVDS_OK)
#endif
    {
        uint8_t counter;

        // generate a new IRK
        for (counter = 0; counter < KEY_LEN; counter++)
        {
            app_env.loc_irk[counter] = (uint8_t)co_rand_word();
        }
#if (NVDS_SUPPORT)
        // Store the generated value in NVDS
        if (nvds_put(NVDS_TAG_LOC_IRK, KEY_LEN, (uint8_t *)&app_env.loc_irk) != NVDS_OK)
        {
            ASSERT_INFO(0, NVDS_TAG_LOC_IRK, 0);
        }
#endif
    }

    /*------------------------------------------------------
     * INITIALIZE ALL MODULES
     *------------------------------------------------------*/

    // Device Information Module
    app_dis_init();

    // Battery Module
    app_batt_init();

#if BLE_OADS_SERVER
    app_oads_init();
#endif

#if BLE_CUSTOM_SERVER
    app_custom_init();
#endif

#ifdef UART_TP_MODE
    app_midi_init();
#endif
}

bool appm_add_svc(void)
{
    // Indicate if more services need to be added in the database
    bool more_svc = false;

    // Check if another should be added in the database
    if (app_env.next_svc != APPM_SVC_LIST_STOP)
    {
        ASSERT_INFO(appm_add_svc_func_list[app_env.next_svc] != NULL, app_env.next_svc, 1);

        // Call the function used to add the required service
        appm_add_svc_func_list[app_env.next_svc]();

        // Select following service to add
        app_env.next_svc++;
        more_svc = true;
    }

    return more_svc;
}

/*�豸�����Ͽ����Ӻ���*/
void appm_disconnect(void)
{
    struct gapc_disconnect_cmd *cmd = KE_MSG_ALLOC(GAPC_DISCONNECT_CMD,
                                                   KE_BUILD_ID(TASK_GAPC, app_env.conidx), TASK_APP,
                                                   gapc_disconnect_cmd);

    cmd->operation = GAPC_DISCONNECT;
    cmd->reason = CO_ERROR_REMOTE_USER_TERM_CON;
    UART_PRINTF("%s\r\n", __func__);
    // Send the message
    ke_msg_send(cmd);
}

void appm_start_advertising(void)
{
    // Check if the advertising procedure is already is progress
    if (ke_state_get(TASK_APP) == APPM_READY)
    {
        // Prepare the GAPM_START_ADVERTISE_CMD message
        struct gapm_start_advertise_cmd *cmd = KE_MSG_ALLOC(GAPM_START_ADVERTISE_CMD,
                                                            TASK_GAPM, TASK_APP,
                                                            gapm_start_advertise_cmd);

        cmd->op.addr_src = GAPM_STATIC_ADDR;
        cmd->channel_map = APP_ADV_CHMAP;
        cmd->intv_min = APP_ADV_INT_MIN;
        cmd->intv_max = APP_ADV_INT_MAX;
        cmd->op.code = GAPM_ADV_UNDIRECT;

        cmd->info.host.mode = GAP_GEN_DISCOVERABLE;

        /*-----------------------------------------------------------------------------------
         * Set the Advertising Data and the Scan Response Data
         *---------------------------------------------------------------------------------*/
        // Flag value is set by the GAP
        cmd->info.host.adv_data_len = ADV_DATA_LEN;
        cmd->info.host.scan_rsp_data_len = SCAN_RSP_DATA_LEN;

#if (NVDS_SUPPORT)
        // Advertising Data
        if (nvds_get(NVDS_TAG_APP_BLE_ADV_DATA, &cmd->info.host.adv_data_len,
                     &cmd->info.host.adv_data[0]) != NVDS_OK)
#endif
        {
            // cmd->info.host.adv_data_len = 0;

            cmd->info.host.adv_data[0] = 2; // Length of ad type flags
            cmd->info.host.adv_data[1] = GAP_AD_TYPE_FLAGS;
            cmd->info.host.adv_data[2] = GAP_BR_EDR_NOT_SUPPORTED;
            // set mode in ad_type
            switch (cmd->info.host.mode)
            {
            // General discoverable mode
            case GAP_GEN_DISCOVERABLE:
            {
                cmd->info.host.adv_data[2] |= GAP_LE_GEN_DISCOVERABLE_FLG;
            }
            break;
            // Limited discoverable mode
            case GAP_LIM_DISCOVERABLE:
            {
                cmd->info.host.adv_data[2] |= GAP_LE_LIM_DISCOVERABLE_FLG;
            }
            break;
            default:
                break; // do nothing
            }
            cmd->info.host.adv_data_len = 3;
            {
                uint8_t uuid_midi_adv_data[18];
#if 0
                //Add list of UUID and appearance						
                memcpy(&cmd->info.host.adv_data[cmd->info.host.adv_data_len],
                       APP_FFF0_ADV_DATA_UUID, APP_FFF0_ADV_DATA_UUID_LEN);
                cmd->info.host.adv_data_len += APP_FFF0_ADV_DATA_UUID_LEN;
#else
                uuid_midi_adv_data[0] = 0x11;
                uuid_midi_adv_data[1] = 0x07; // declare 128-bit uuid
                uuid_midi_adv_data[2] = 0x00;
                uuid_midi_adv_data[3] = 0xC7; // 2018/12/25 fix3
                uuid_midi_adv_data[4] = 0xC4;
                uuid_midi_adv_data[5] = 0x4E;
                uuid_midi_adv_data[6] = 0xE3;
                uuid_midi_adv_data[7] = 0x6C;
                uuid_midi_adv_data[8] = 0x51;
                uuid_midi_adv_data[9] = 0xA7;
                uuid_midi_adv_data[10] = 0x33;
                uuid_midi_adv_data[11] = 0x4B;
                uuid_midi_adv_data[12] = 0xE8;
                uuid_midi_adv_data[13] = 0xED;
                uuid_midi_adv_data[14] = 0x5A;
                uuid_midi_adv_data[15] = 0x0E;
                uuid_midi_adv_data[16] = 0xB8;
                uuid_midi_adv_data[17] = 0x03;
                memcpy(&cmd->info.host.adv_data[cmd->info.host.adv_data_len],
                       uuid_midi_adv_data, APP_MIDI_ADV_DATA_UUID_LEN);
                cmd->info.host.adv_data_len += APP_MIDI_ADV_DATA_UUID_LEN;
#endif
            }
        }

        //  Device Name Length
        uint8_t device_name_length;
        uint8_t device_name_avail_space;
        uint8_t device_name_temp_buf[APP_DEVICE_NAME_LENGTH_MAX];

        // Get remaining space in the Advertising Data - 2 bytes are used for name length/flag
        device_name_avail_space = ADV_DATA_LEN - cmd->info.host.adv_data_len - 2;

        // Check if data can be added to the Advertising data
        if (device_name_avail_space > 2)
        {

            device_name_length = NVDS_LEN_DEVICE_NAME;
#if (NVDS_SUPPORT)
            if (nvds_get(NVDS_TAG_DEVICE_NAME, &device_name_length,
                         &device_name_temp_buf[0]) != NVDS_OK)
#endif
            {
                device_name_length = strlen(APP_DFLT_DEVICE_NAME);
                // Get default Device Name (No name if not enough space)
                memcpy(&device_name_temp_buf[0], APP_DFLT_DEVICE_NAME, device_name_length);
                UART_PRINTF("dev_name :%s\r\n", device_name_temp_buf);
            }

            if (device_name_length > 0)
            {
                // Check available space
                device_name_length = co_min(device_name_length, device_name_avail_space);
                cmd->info.host.adv_data[cmd->info.host.adv_data_len] = device_name_length + 1;

                // Fill Device Name Flag
                cmd->info.host.adv_data[cmd->info.host.adv_data_len + 1] = '\x09';
                // Copy device name
                memcpy(&cmd->info.host.adv_data[cmd->info.host.adv_data_len + 2],
                       device_name_temp_buf, device_name_length);

                // Update Advertising Data Length
                cmd->info.host.adv_data_len += (device_name_length + 2);
            }
        }
#if (NVDS_SUPPORT)
        // Scan Response Data
        if (nvds_get(NVDS_TAG_APP_BLE_SCAN_RESP_DATA, &cmd->info.host.scan_rsp_data_len,
                     &cmd->info.host.scan_rsp_data[0]) != NVDS_OK)
#endif
        {
            cmd->info.host.scan_rsp_data_len = 0;
            // Get from APP_DFLT_DEVICE_NAME
            //
            scan_response_data[0] = strlen(APP_DFLT_DEVICE_NAME) + 2;
            scan_response_data[1] = '\x08';
            memcpy(&scan_response_data[2], APP_DFLT_DEVICE_NAME, strlen(APP_DFLT_DEVICE_NAME));
            memcpy(&cmd->info.host.scan_rsp_data[cmd->info.host.scan_rsp_data_len],
                   APP_SCNRSP_DATA, APP_SCNRSP_DATA_LEN);
            cmd->info.host.scan_rsp_data_len += APP_SCNRSP_DATA_LEN;
        }
        appm_set_dev_name(APP_DFLT_DEVICE_NAME, strlen(APP_DFLT_DEVICE_NAME));
        // Send the message
        ke_msg_send(cmd);
        UART_PRINTF("appm start advertising\r\n");
        // wdt_disable();
        wdt_enable(0x3fff);

        // Set the state of the task to APPM_ADVERTISING
        ke_state_set(TASK_APP, APPM_ADVERTISING);

        // ke_timer_set(APP_PERIOD_TIMER, TASK_APP, 1500);
    }

    // else ignore the request
}

/* �豸����ֹͣ�㲥����*/
void appm_stop_advertising(void)
{
    if (ke_state_get(TASK_APP) == APPM_ADVERTISING)
    {
        UART_PRINTF("%s\r\n", __func__);
        // Go in ready state
        ke_state_set(TASK_APP, APPM_READY);

        // Prepare the GAPM_CANCEL_CMD message
        struct gapm_cancel_cmd *cmd = KE_MSG_ALLOC(GAPM_CANCEL_CMD,
                                                   TASK_GAPM, TASK_APP,
                                                   gapm_cancel_cmd);
        cmd->operation = GAPM_CANCEL;

        // Send the message
        ke_msg_send(cmd);

        wdt_disable_flag = 1;
    }
    // else ignore the request
}

void appm_update_param(struct gapc_conn_param *conn_param)
{
    // Prepare the GAPC_PARAM_UPDATE_CMD message
    struct gapc_param_update_cmd *cmd = KE_MSG_ALLOC(GAPC_PARAM_UPDATE_CMD,
                                                     KE_BUILD_ID(TASK_GAPC, app_env.conidx), TASK_APP,
                                                     gapc_param_update_cmd);

    cmd->operation = GAPC_UPDATE_PARAMS;
    cmd->intv_min = conn_param->intv_min;
    cmd->intv_max = conn_param->intv_max;
    cmd->latency = conn_param->latency;
    cmd->time_out = conn_param->time_out;

    // not used by a slave device
    cmd->ce_len_min = 0xFFFF;
    cmd->ce_len_max = 0xFFFF;

    UART_PRINTF("intv_min = %d,intv_max = %d,latency = %d,time_out = %d\r\n",
                cmd->intv_min, cmd->intv_max, cmd->latency, cmd->time_out);

    // Send the message
    ke_msg_send(cmd);
}
uint8_t appm_set_dev_name(char *name, uint8_t len)
{
    app_env.dev_name_len = len;
    // copy name to provided pointer
    memcpy(app_env.dev_name, name, len);
    // return name length
    return app_env.dev_name_len;
}

uint8_t appm_get_dev_name(uint8_t *name)
{
    // copy name to provided pointer
    memcpy(name, app_env.dev_name, app_env.dev_name_len);
    // return name length
    return app_env.dev_name_len;
}
// user define
uint8_t wait_to_start_advertising = 0;
uint8_t BACT_STATE = BACT_STATE_ACTIVE;
uint8_t rxdata_buffer[128];
uint8_t rxdata_buffer_len = 0;
uint8_t AT_cmd_buf[64];
uint8_t recv_AT_cmd_idx = 0;
uint8_t AT_cmd_len = 0;
#define UW_ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
#define MIN(a, b) ((a) > (b) ? (b) : (a))
uint8_t cmd_rsp[AT_RESULT_MAX] = {
    AT_RESULT_OK0,    // vs AT_RESULT_AT
    AT_RESULT_OK1,    // vs AT_RESULT_BAUD1
    AT_RESULT_OK2,    // vs AT_RESULT_BAUD2
    AT_RESULT_OK3,    // vs AT_RESULT_BAUD3
    AT_RESULT_OK4,    // vs AT_RESULT_BAUD4
    AT_RESULT_OK5,    // vs AT_RESULT_BAUD5
    AT_RESULT_OK1,    // vs AT_RESULT_BAUDS1
    AT_RESULT_OK2,    // vs AT_RESULT_BAUDS2
    AT_RESULT_OK3,    // vs AT_RESULT_BAUDS3
    AT_RESULT_OK4,    // vs AT_RESULT_BAUDS4
    AT_RESULT_OK5,    // vs AT_RESULT_BAUDS5
    AT_RESULT_OKDATA, // vs AT_RESULT_NAME
    AT_RESULT_OKDATA, // vs AT_RESULT_NAMES
    AT_RESULT_OKDATA, // vs AT_RESULT_SUUID
    AT_RESULT_OKDATA, // vs AT_RESULT_SUUIDS
    AT_RESULT_OKDATA, // vs AT_RESULT_WUUID
    AT_RESULT_OKDATA, // vs AT_RESULT_WUUIDS
    AT_RESULT_OKDATA, // vs AT_RESULT_NUUID
    AT_RESULT_OKDATA, // vs AT_RESULT_NUUIDS
    AT_RESULT_ADVOK1, // vs AT_RESULT_ADVI1
    AT_RESULT_ADVOK2, // vs AT_RESULT_ADVI2
    AT_RESULT_ADVOK3, // vs AT_RESULT_ADVI3
    AT_RESULT_ADVOK4, // vs AT_RESULT_ADVI4
    AT_RESULT_ADVOK5, // vs AT_RESULT_ADVI5
    AT_RESULT_ADVOK1, // vs AT_RESULT_ADVIS1
    AT_RESULT_ADVOK2, // vs AT_RESULT_ADVIS2
    AT_RESULT_ADVOK3, // vs AT_RESULT_ADVIS3
    AT_RESULT_ADVOK4, // vs AT_RESULT_ADVIS4
    AT_RESULT_ADVOK5, // vs AT_RESULT_ADVIS5
    AT_RESULT_OKDATA, // vs AT_RESULT_RFLASH
    AT_RESULT_OKDATA, // vs AT_RESULT_WFLASH
};
static const char *get_at_rsp(uint8_t code)
{
    uint8_t i;
    for (i = 0; i < UW_ARRAY_SIZE(at_prefix_rsp); i++)
    {
        if (at_prefix_rsp[i].code == code)
        {
            return at_prefix_rsp[i].prefix;
        }
    }

    return NULL;
}
static at_prefix_t *at_result_to_prefix(char *buf, int32_t len)
{
    uint16_t i;
    at_prefix_t *p;
    // DECLARE_FNAME("at_result_parse_and_store");

    for (i = 0, p = (at_prefix_t *)at_prefix_req; i < UW_ARRAY_SIZE(at_prefix_req); i++, p++)
    {
        if (len >= p->prefix_len &&
            strncmp(buf, p->prefix, MIN(len, p->prefix_len)) == 0)
        {
            return p;
        }
    }

    // DBG_W(DBT_HFP, ("%s: unknown AT result\n", FNAME));
    // DBG_HEX_DUMP(DBT_HFP, DL_WARNING, buf, len, 0);
    return NULL;
}

uint8_t Char_to_Hex(uint8_t input_char)
{
    if ((input_char <= 'F') && (input_char >= 'A'))
    {
        return ((input_char - 'A') + 10);
    }
    else if ((input_char <= 'f') && (input_char >= 'a'))
    {
        return ((input_char - 'a') + 10);
    }
    else if ((input_char <= '9') && (input_char >= '0'))
    {
        return (input_char - '0');
    }
    return 0;
}

extern int GAPM_reset(void);
#if 1
void app_user_init(void)
{
    NVR_extMCU_USED_FLASH_AREA_info_get();
    //    NVR_BK3432_USED_FLASH_AREA_info_get();
#if 1
    load_parameter_from_BK3432_USED_FLASH_AREA();
/*
switch(Sram_Baudrate)
{
    case 1 :
        uart_init(9600);
        UART_PRINTF("baud :9600\r\n");
    break;
    case 2 :
        uart_init(19200);
        UART_PRINTF("baud :19200\r\n");
    break;
    case 3 :
        uart_init(38400);
        UART_PRINTF("baud :38400\r\n");
    break;
    case 4 :
        uart_init(57600);
        UART_PRINTF("baud :57600\r\n");
    break;
    case 5 :
        uart_init(115200);
        UART_PRINTF("baud :115200\r\n");
    break;
    default :
        uart_init(9600);
        UART_PRINTF("baud :9600\r\n");
    break;
}
*/
#endif
}
#endif
void update_SRAM_name()
{
}
void app_user_entry(void)
{
    uint8_t without_prefix_len;

    // GPIO_int_enable();
    if (BACT_STATE != BACT_STATE_DEEPSLEEP)
    {

        if (ke_state_get(TASK_APP) == APPM_READY)
        {
            UART_PRINTF("start advertising\r\n");
            appm_start_advertising();
        }
        else if (ke_state_get(TASK_APP) == APPM_CONNECTED)
        {
            if (uart_rx_done == 1 && uart_tp_mode_flag == true)
            {
                if (rxdata_buffer_len > 128)
                {
                    rxdata_buffer_len = 128;
                }

                if (strcmp((char *)rxdata_buffer, "AT+TPMODE0\r\n") != 0 && strcmp((char *)rxdata_buffer, "AT+TPMODE1\r\n") != 0)
                {
#ifdef RXDATA_DEBUG
                    uint8_t i;
                    uart2_printf("[%s][%u]", __func__, __LINE__);
                    for (i = 0; i < rxdata_buffer_len; i++)
                    {
                        UART2_PRINTF("%02x ", rxdata_buffer[i]);
                    }
                    uart2_printf("\r\n");
#endif
                    app_midi_send_payload(rxdata_buffer, rxdata_buffer_len);

                    rxdata_buffer_len = 0;
                    uart_rx_done = 0;
                }
            }
        }

        if (uart_rx_done == 1)
        {
            // uint8_t baud_change = 0;
            uint8_t len;
            uint8_t rsp_code;
            // uint8_t idx;
            extern uint8_t rxdata_buffer_len;
            at_prefix_t *prefix_cmd;
            uint8_t w_flash_buf[2];
            // len = strlen((char*)rxdata_buffer);
            len = rxdata_buffer_len;
            rxdata_buffer_len = 0;
            if (rxdata_buffer[len - 1] == '\n')
            {
                // AT command finish
                // UART_PRINTF("finish\r\n");
                memcpy(&AT_cmd_buf[recv_AT_cmd_idx], rxdata_buffer, len);
                // UART_PRINTF("%s\r\n",AT_cmd_buf);
                AT_cmd_len += len;
                recv_AT_cmd_idx = 0;
            }
            else
            {
                // command not finish
                memcpy(&AT_cmd_buf[recv_AT_cmd_idx], rxdata_buffer, len);
                recv_AT_cmd_idx = len;
                AT_cmd_len += len;
                uart_rx_done = 0;
                // UART_PRINTF("not finish\r\n");
                return;
            }

            // dispatch AT-COMMAND
            prefix_cmd = at_result_to_prefix((char *)AT_cmd_buf, AT_cmd_len);
            uart_rx_done = 0;
            without_prefix_len = AT_cmd_len - prefix_cmd->prefix_len;
            //                  UART_PRINTF("len : %d\r\n",without_prefix_len);//len are include "\r\n" char
            if (prefix_cmd == NULL)
            {
#ifdef used_BK3432_MCU
                UART_SEND_DATA("@");
#endif
                UART_SEND_DATA("ERROR\r\n");
            }
            else if (prefix_cmd != NULL)
            {
                rsp_code = cmd_rsp[prefix_cmd->code];
                switch (prefix_cmd->code)
                {
                case AT_RESULT_AT:
                    // do nothing

#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    break;
                case AT_RESULT_BAUD1:
                    UART_PRINTF("recv AT_RESULT_BAUD1\r\n");
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    cpu_delay(15);
                    uart_init(9600);
                    w_flash_buf[0] = 1;
                    save_parameter_to_BK3432_USED_FLASH_AREA(TAG_BAUD, w_flash_buf);

                    break;
                case AT_RESULT_BAUD2:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    cpu_delay(15);
                    w_flash_buf[0] = 2;
                    save_parameter_to_BK3432_USED_FLASH_AREA(TAG_BAUD, w_flash_buf);
                    uart_init(19200);

                    break;
                case AT_RESULT_BAUD3:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    cpu_delay(15);
                    w_flash_buf[0] = 3;
                    save_parameter_to_BK3432_USED_FLASH_AREA(TAG_BAUD, w_flash_buf);
                    uart_init(38400);

                    break;
                case AT_RESULT_BAUD4:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code)); // 2018/12/25 fix4,use old baudrate send response and change new baudrate
                    cpu_delay(15);
                    w_flash_buf[0] = 4;
                    save_parameter_to_BK3432_USED_FLASH_AREA(TAG_BAUD, w_flash_buf);
                    uart_init(57600);
                    break;
                case AT_RESULT_BAUD5:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code)); // 2018/12/25 fix4,use old baudrate send response and change new baudrate
                    cpu_delay(15);
                    w_flash_buf[0] = 5;
                    save_parameter_to_BK3432_USED_FLASH_AREA(TAG_BAUD, w_flash_buf);
                    uart_init(115200);
                    break;
                case AT_RESULT_BAUDS1:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code)); // 2018/12/25 fix4,use old baudrate send response and change new baudrate
                    cpu_delay(15);
                    uart_init(9600);
                    break;
                case AT_RESULT_BAUDS2:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code)); // 2018/12/25 fix4,use old baudrate send response and change new baudrate
                    cpu_delay(15);
                    uart_init(19200);
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    break;
                case AT_RESULT_BAUDS3:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code)); // 2018/12/25 fix4,use old baudrate send response and change new baudrate
                    cpu_delay(15);
                    uart_init(38400);
                    break;
                case AT_RESULT_BAUDS4:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code)); // 2018/12/25 fix4,use old baudrate send response and change new baudrate
                    cpu_delay(15);
                    uart_init(57600);
                    break;
                case AT_RESULT_BAUDS5:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code)); // 2018/12/25 fix4,use old baudrate send response and change new baudrate
                    cpu_delay(15);
                    uart_init(115200);
                    break;
                case AT_RESULT_NAME:
                case AT_RESULT_NAMES:
                    cpu_delay(15);
                    memset(Sram_Device_name, 0, 20);
                    memcpy(APP_DFLT_DEVICE_NAME, AT_cmd_buf + prefix_cmd->prefix_len, (AT_cmd_len - prefix_cmd->prefix_len));
                    appm_set_dev_name(APP_DFLT_DEVICE_NAME, (AT_cmd_len - prefix_cmd->prefix_len));
                    AT_cmd_len = 0;

                    if (prefix_cmd->code == AT_RESULT_NAME)
                    {
                        save_parameter_to_BK3432_USED_FLASH_AREA(TAG_NAME, (uint8_t *)APP_DFLT_DEVICE_NAME);
                    }
                    {
                        char buffer_A[64] = {0};

#ifdef used_BK3432_MCU
                        UART_SEND_DATA("@");
#endif
                        UART_SEND_DATA("%s", get_at_rsp(rsp_code));
                        strncpy(buffer_A, (char *)(AT_cmd_buf + prefix_cmd->prefix_len), (without_prefix_len - 2));
                        UART_SEND_DATA("%s\r\n", buffer_A);
                    }
                    // UART_PRINTF("name : %s\r\n",APP_DFLT_DEVICE_NAME);
                    // cpu_delay(15);
                    // memcpy(APP_DFLT_DEVICE_NAME,AT_cmd_buf+prefix_cmd->prefix_len,(AT_cmd_len-prefix_cmd->prefix_len));
                    // AT_cmd_len = 0;

                    // UART_PRINTF("name : %s\r\n",APP_DFLT_DEVICE_NAME);
                    break;
                case AT_RESULT_SUUID:  // for FLASH
                case AT_RESULT_SUUIDS: // for SRAM
                {
                    AT_cmd_len = 0;
#if 1
                    uint8_t i;
                    uint16_t number = 0;
                    cpu_delay(15);
                    UART_PRINTF("without_prefix_len : %d\r\n", without_prefix_len);
                    UART_PRINTF("prefix_cmd->prefix_len : %d\r\n", prefix_cmd->prefix_len);
                    UART_PRINTF("UUID :");

                    without_prefix_len -= 2; // without "\r\n"

                    Sram_FFF0_Service = 0;
                    for (i = 0; i < without_prefix_len; i++)
                    {
                        number = Char_to_Hex(*(AT_cmd_buf + (prefix_cmd->prefix_len + i)));
                        Sram_FFF0_Service |= (number << (12 - i * 4));
                        UART_PRINTF(" %x", number);
                    }
                    Sram_FFFx_load_from_flash &= ~FFF0_LOAD_MASK;
                    if (prefix_cmd->code == AT_RESULT_SUUID)
                    {
                        Sram_FFFx_load_from_flash |= FFF0_LOAD_MASK;
                        w_flash_buf[0] = Sram_FFF0_Service >> 8;
                        w_flash_buf[1] = Sram_FFF0_Service & 0x00FF;
                        save_parameter_to_BK3432_USED_FLASH_AREA(TAG_FFF0, w_flash_buf);
                    }
                    {
                        char buffer_A[64] = {0};
#ifdef used_BK3432_MCU
                        UART_SEND_DATA("@");
#endif
                        UART_SEND_DATA("%s", get_at_rsp(rsp_code));
                        strncpy(buffer_A, (char *)(AT_cmd_buf + prefix_cmd->prefix_len), without_prefix_len);
                        UART_SEND_DATA("%s\r\n", buffer_A);
                    }
                    UART_PRINTF(" %x ", Sram_FFF0_Service);
                    UART_PRINTF("\r\n");
#endif
                    // Sram_FFF0_Service = ((_interger_number&0xFF00) >> 8);
                    GAPM_reset();
                    // memcpy(Sram_FFF0_Service,AT_cmd_buf+prefix_cmd->prefix_len,(AT_cmd_len-prefix_cmd->prefix_len));
                }
                break;
                // case AT_RESULT_SUUIDS :
                // break;
                case AT_RESULT_NUUID:
                case AT_RESULT_NUUIDS:
                {
                    AT_cmd_len = 0;

                    uint8_t i;
                    uint16_t number = 0;
                    cpu_delay(15);
                    UART_PRINTF("UUID :");
                    without_prefix_len -= 2;    // without "\r\n"
                    Sram_FFF1_Char_Service = 0; // Notify
                    for (i = 0; i < without_prefix_len; i++)
                    {
                        number = Char_to_Hex(*(AT_cmd_buf + (prefix_cmd->prefix_len + i)));
                        Sram_FFF1_Char_Service |= (number << (12 - i * 4));
                        UART_PRINTF(" %x", number);
                    }
                    Sram_FFFx_load_from_flash &= ~FFF1_LOAD_MASK;
                    if (prefix_cmd->code == AT_RESULT_NUUID)
                    {
                        Sram_FFFx_load_from_flash |= FFF1_LOAD_MASK;
                        w_flash_buf[0] = Sram_FFF1_Char_Service >> 8;
                        w_flash_buf[1] = Sram_FFF1_Char_Service & 0x00FF;
                        save_parameter_to_BK3432_USED_FLASH_AREA(TAG_FFF1, w_flash_buf);
                    }
                    {
                        char buffer_A[64] = {0};

#ifdef used_BK3432_MCU
                        UART_SEND_DATA("@");
#endif
                        UART_SEND_DATA("%s", get_at_rsp(rsp_code));
                        strncpy(buffer_A, (char *)(AT_cmd_buf + prefix_cmd->prefix_len), without_prefix_len);
                        UART_SEND_DATA("%s\r\n", buffer_A);
                    }
                    UART_PRINTF(" %x ", Sram_FFF1_Char_Service);
                    UART_PRINTF("\r\n");

                    // Sram_FFF0_Service = ((_interger_number&0xFF00) >> 8);
                    GAPM_reset();
                    // memcpy(Sram_FFF0_Service,AT_cmd_buf+prefix_cmd->prefix_len,(AT_cmd_len-prefix_cmd->prefix_len));
                }
                break;
                // case AT_RESULT_WUUIDS :
                // break;
                case AT_RESULT_WUUID:
                case AT_RESULT_WUUIDS:
                {
                    AT_cmd_len = 0;
#if 1
                    uint8_t i;
                    uint16_t number = 0;
                    cpu_delay(15);
                    UART_PRINTF("UUID :");
                    without_prefix_len -= 2;    // without "\r\n"
                    Sram_FFF2_Char_Service = 0; // Write without response
                    for (i = 0; i < without_prefix_len; i++)
                    {
                        number = Char_to_Hex(*(AT_cmd_buf + (prefix_cmd->prefix_len + i)));
                        Sram_FFF2_Char_Service |= (number << (12 - i * 4));
                        UART_PRINTF(" %x", number);
                    }
                    Sram_FFFx_load_from_flash &= ~FFF2_LOAD_MASK;

                    if (prefix_cmd->code == AT_RESULT_WUUID)
                    {
                        Sram_FFFx_load_from_flash |= FFF2_LOAD_MASK;
                        w_flash_buf[0] = Sram_FFF2_Char_Service >> 8;
                        w_flash_buf[1] = Sram_FFF2_Char_Service & 0x00FF;
                        save_parameter_to_BK3432_USED_FLASH_AREA(TAG_FFF2, w_flash_buf);
                    }
                    {
                        char buffer_A[64] = {0};

#ifdef used_BK3432_MCU
                        UART_SEND_DATA("@");
#endif
                        UART_SEND_DATA("%s", get_at_rsp(rsp_code));
                        strncpy(buffer_A, (char *)(AT_cmd_buf + prefix_cmd->prefix_len), without_prefix_len);
                        UART_SEND_DATA("%s\r\n", buffer_A);
                    }
                    UART_PRINTF(" %x ", Sram_FFF2_Char_Service);
                    UART_PRINTF("\r\n");
#endif
                    // Sram_FFF0_Service = ((_interger_number&0xFF00) >> 8);
                    GAPM_reset();
                    // memcpy(Sram_FFF0_Service,AT_cmd_buf+prefix_cmd->prefix_len,(AT_cmd_len-prefix_cmd->prefix_len));
                }
                break;
                case AT_RESULT_ADVI1:
                case AT_RESULT_ADVIS1:
                    Sram_Adv_interval = 160;
                    if (prefix_cmd->code == AT_RESULT_ADVI1)
                    {
                        w_flash_buf[0] = 1;
                        save_parameter_to_BK3432_USED_FLASH_AREA(TAG_ADV, w_flash_buf);
                    }
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif

                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    break;
                case AT_RESULT_ADVI2:
                case AT_RESULT_ADVIS2:
                    Sram_Adv_interval = 400;
                    if (prefix_cmd->code == AT_RESULT_ADVI2)
                    {
                        w_flash_buf[0] = 2;
                        save_parameter_to_BK3432_USED_FLASH_AREA(TAG_ADV, w_flash_buf);
                    }
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    break;
                case AT_RESULT_ADVI3:
                case AT_RESULT_ADVIS3:
                    Sram_Adv_interval = 800;
                    if (prefix_cmd->code == AT_RESULT_ADVI3)
                    {
                        w_flash_buf[0] = 3;
                        save_parameter_to_BK3432_USED_FLASH_AREA(TAG_ADV, w_flash_buf);
                    }
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    break;
                case AT_RESULT_ADVI4:
                case AT_RESULT_ADVIS4:
                    Sram_Adv_interval = 2560;
                    if (prefix_cmd->code == AT_RESULT_ADVI4)
                    {
                        w_flash_buf[0] = 4;
                        save_parameter_to_BK3432_USED_FLASH_AREA(TAG_ADV, w_flash_buf);
                    }
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    break;
                case AT_RESULT_ADVI5:
                case AT_RESULT_ADVIS5:
                    Sram_Adv_interval = 5120;
                    if (prefix_cmd->code == AT_RESULT_ADVI5)
                    {
                        w_flash_buf[0] = 5;
                        save_parameter_to_BK3432_USED_FLASH_AREA(TAG_ADV, w_flash_buf);
                    }
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s\r\n", get_at_rsp(rsp_code));
                    break;
                case AT_RESULT_RFLASH:
                {
                    uint8_t i, addr = 0, number, data;

                    for (i = 0; i < 2; i++)
                    {
                        number = Char_to_Hex(*(AT_cmd_buf + (prefix_cmd->prefix_len + i)));
                        addr |= (number << (4 - i * 4));
                    }
                    // UART_PRINTF("addr: %x",addr);
                    data = RFLASH_process(addr, extMCU_USED_FLASH_AREA);

#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s", get_at_rsp(rsp_code));
                    if (addr == 0x00)
                    {
                        UART_SEND_DATA("00");
                    }
                    else
                    {
                        UART_SEND_DATA("%02x", addr);
                    }
                    if (data == 0x00)
                    {
                        UART_SEND_DATA("00\r\n");
                    }
                    else
                    {
                        UART_SEND_DATA("%02x\r\n", data);
                    }
                }
                break;
                case AT_RESULT_WFLASH:
                {
                    uint8_t i, addr = 0, number, data = 0, rdata;

                    for (i = 0; i < 2; i++)
                    {
                        number = Char_to_Hex(AT_cmd_buf[prefix_cmd->prefix_len + i]);
                        addr |= (number << (4 - i * 4));
                    }

                    for (i = 2; i < 4; i++)
                    {
                        number = Char_to_Hex(AT_cmd_buf[prefix_cmd->prefix_len + i]);
                        data |= (number << (4 - ((i - 2) * 4)));
                    }

                    WFLASH_process(addr, data, extMCU_USED_FLASH_AREA);
                    rdata = RFLASH_process(addr, extMCU_USED_FLASH_AREA);
                    UART_PRINTF("rdaddr : %x,rddata : %x\r\n", addr, rdata);
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("%s", get_at_rsp(rsp_code));
                    if (addr == 0x00)
                    {
                        UART_SEND_DATA("00");
                    }
                    else
                    {
                        UART_SEND_DATA("%02x", addr);
                    }
                    if (data == 0x00)
                    {
                        UART_SEND_DATA("00\r\n");
                    }
                    else
                    {
                        UART_SEND_DATA("%02x\r\n", data);
                    }
                    // UART_SEND_DATA("%02x%02x\r\n",addr,data);//2018/12/25 fix5
                }
                break;
                case AT_RESULT_TPMODE0:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("OKTPMODE0\r\n"); // 2018/12/25 fix4,use old baudrate send response and change new baudrate
                    cpu_delay(15);
                    uart_tp_mode_flag = false;
                    break;
                case AT_RESULT_TPMODE1:
#ifdef used_BK3432_MCU
                    UART_SEND_DATA("@");
#endif
                    UART_SEND_DATA("OKTPMODE1\r\n"); // 2018/12/25 fix4,use old baudrate send response and change new baudrate
                    cpu_delay(15);
                    uart_tp_mode_flag = true;
                    break;
                }
            }
            AT_cmd_len = 0;
        }
    }
    else // BACT_STATE == BACT_STATE_DEEPSLEEP
    {
        // ADV_line_change = 0;
        if (ke_state_get(TASK_APP) == APPM_CONNECTED)
        {
            appm_disconnect();
        }
        if (ke_state_get(TASK_APP) == APPM_ADVERTISING)
        {
            appm_stop_advertising();
        }
    }
}

void NVR_extMCU_USED_FLASH_AREA_info_get(void)
{
    uint8_t i;
    uint8_t r_tmp_buffer[4];
    for (i = 0; i < 128; i++)
    {
        flash_read(FLASH_SPACE_TYPE_NVR, extMCU_USED_FLASH_AREA + i, 4, &r_tmp_buffer[0]); // every time read 4 bytes
        // void update_page_info_data_in_SRAM(uint16_t area,uint8_t page_no,uint8_t info,uint8_t data);
        // extMCU_used_sram_data_info[i] = r_tmp_buffer[0];
        update_page_info_data(extMCU_USED_FLASH_AREA, i, r_tmp_buffer[0], r_tmp_buffer[flash_unit_info_transf_data_idx(r_tmp_buffer[0])]);
    }
}
#if 0
void NVR_BK3432_USED_FLASH_AREA_info_get(void)
{
    uint8_t i;
    uint8_t r_tmp_buffer[4];
    for(i = 0;i < 128;i++)
    {
        flash_read(FLASH_SPACE_TYPE_NVR, BK3432_USED_FLASH_AREA+i, 4, &r_tmp_buffer[0]);//every time read 4 bytes
        //void update_page_info_data_in_SRAM(uint16_t area,uint8_t page_no,uint8_t info,uint8_t data);
        //extMCU_used_sram_data_info[i] = r_tmp_buffer[0];
        update_page_info_data(BK3432_USED_FLASH_AREA,i,r_tmp_buffer[0],r_tmp_buffer[flash_unit_info_transf_data_idx(r_tmp_buffer[0])]);
    }
    
}
#endif
uint8_t flash_unit_info_transf_data_idx(uint8_t info)
{
    uint8_t ret_idx = 0, i;
    info = ~info;
    for (i = 0; i < 3; i++)
    {
        if ((info & 0x01) & 0x01)
        {
            ret_idx++;
        }
        info >>= 1;
    }
    // UART_PRINTF("%s,info : %x,ret_idx : %x\r\n",info,ret_idx);
    return ret_idx;
}

uint8_t RFLASH_process(uint8_t address, uint16_t area)
{
    uint8_t data_offset = 0;
    uint8_t r_tmp_buffer[4];
    uint8_t ret_info;
    uint8_t ret_data;

    // address &= 0x7F;//limit into 0~127 bytes
    // check sram_data_unit[i] info
    Get_page_info_data(area, address, &ret_data, &ret_info); // extMCU read SRAM info table,BK3432 used direct read flash info byte
    data_offset = flash_unit_info_transf_data_idx(ret_info);
    // UART_PRINTF(" address : %d, info : %x,data_offset : %d\r\n",address,sram_data_unit[address].info,data_offset);
    if (data_offset != 0)
    {
        flash_read(FLASH_SPACE_TYPE_NVR, area + address, 4, &r_tmp_buffer[0]);
        // UART_PRINTF("RFLASH_process,data : %x%x%x%x\r\n",r_tmp_buffer[0],r_tmp_buffer[1],r_tmp_buffer[2],r_tmp_buffer[3]);
        return r_tmp_buffer[data_offset];
    }
    else // no data in the flash
    {
        return 0xff;
    }
}
void Get_page_info_data(uint16_t area, uint8_t page_no, uint8_t *data, uint8_t *info)
{

    if (area == extMCU_USED_FLASH_AREA)
    {
        *info = extMCU_used_sram_data_info[page_no].info;
        *data = extMCU_used_sram_data_info[page_no].data;
    }
    else if (area == BK3432_USED_FLASH_AREA)
    {
        *info = Get_page_info_data_in_FLASH(area, page_no);
        //*info = BK3432_used_sram_data_info[page_no].info;
        *data = 0xFF;
    }
}
void update_page_info_data(uint16_t area, uint8_t page_no, uint8_t info, uint8_t data)
{
    if (area == extMCU_USED_FLASH_AREA)
    {
        // UART_PRINTF("%s,info : %x,data : %x",__func__,info,data);
        extMCU_used_sram_data_info[page_no].info = info;
        extMCU_used_sram_data_info[page_no].data = data;
    }
    else if (area == BK3432_USED_FLASH_AREA)
    {
        return;
        // BK3432_used_sram_data_info[page_no].info = info;
        // BK3432_used_sram_data_info[page_no].data = data;
    }
}
void WFLASH_process(uint8_t address, uint8_t data, uint16_t area)
{
    uint8_t data_offset = 0;
    uint8_t w_tmp_buffer[4] = {0xFF, 0xFF, 0xFF, 0xFF};
    // uint8_t r_tmp_buffer[4] = {0xFF,0xFF,0xFF,0xFF};
    uint8_t idx;
    uint8_t ret_info;
    uint8_t ret_data;

    // address &= 0x7F;//limit into 0~127 bytes
    // check sram_data_unit[i] info
    UART_PRINTF("%s\r\n", __func__);
    Get_page_info_data(area, address, &ret_data, &ret_info); // extMCU read SRAM info table,BK3432 used direct read flash info byte
    data_offset = flash_unit_info_transf_data_idx(ret_info);
    UART_PRINTF("data_offset : %d,address : %x\r\n", data_offset, address);
    if (data_offset != 0) // flash have data
    {
        if (data_offset == 3) // flash no space to write
        {
            UART_PRINTF("flash no space\r\n");
            // copy to temp buffer;
            for (idx = 0; idx < 128; idx++)
            {
                flash_read(FLASH_SPACE_TYPE_NVR, area + idx, 4, &w_tmp_buffer[0]);

                Get_page_info_data(area, address, &ret_data, &ret_info);
                data_offset = flash_unit_info_transf_data_idx(ret_info);
                if (data_offset != 0) // flash have data
                {
                    // check info data if data exist ,change info value to 0xFE
                    // info = Get_page_info_in_SRAM(area,address);
                    // Get_page_info_data_in_SRAM(area,idx,&ret_data,&ret_info);

                    // data_tmp = w_tmp_buffer[data_offset];

                    update_page_info_data(area, idx, 0xFE, w_tmp_buffer[data_offset]); // //copy to temp buffer;
                                                                                       // sram_data_unit[idx].info = 0xFE;
                    // sram_data_unit[idx].data = w_tmp_buffer[data_offset];//flash have data
                }
                else // this address do not have data
                {

                    // sram_data_unit[idx].info = 0xFF;
                    // sram_data_unit[idx].data = 0xFF;
                    update_page_info_data(area, idx, 0xFF, 0xFF);
                }
            }
            flash_erase(FLASH_SPACE_TYPE_NVR, area, 0x200); // erase NVR 0x8000
            for (idx = 0; idx < 128; idx++)
            {
                update_page_info_data(area, idx, 0xFF, 0xFF);
            }
            update_page_info_data(area, address, 0xFE, data); // first,write to sram buffer
            // sram_data_unit[address].info = 0xFE;//update user write data
            // sram_data_unit[address].data = data;
            // write back to flash
            for (idx = 0; idx < 128; idx++)
            {
                // byte assembler to unit(4bytes)
                Get_page_info_data(area, idx, &ret_data, &ret_info);
                if (ret_info == 0xFF)
                {
                    // keep info is 0xFF and do not write data back
                    continue; // do not write to flash
                }
                else
                {
                    // update_page_info_data_in_SRAM(area,address,ret_data,ret_info);
                    w_tmp_buffer[0] = ret_info; // sram_data_unit[idx].info;
                    w_tmp_buffer[1] = ret_data; // sram_data_unit[idx].data;
                    w_tmp_buffer[2] = w_tmp_buffer[3] = 0xFF;
                    flash_write(FLASH_SPACE_TYPE_NVR, area + idx, 4, &w_tmp_buffer[0]);
                }
            }
        }
        else // data_offset == 1,2
        {
            flash_read(FLASH_SPACE_TYPE_NVR, area + address, 4, &w_tmp_buffer[0]);
            if (data_offset == 1)
            {
                ret_info &= 0xFC;
                update_page_info_data(area, address, ret_info, data);
                // sram_data_unit[address].info &= 0xFC;//update info from offset 1 to 2
            }
            else
            {
                ret_info &= 0xF8;
                update_page_info_data(area, address, ret_info, data);
                // sram_data_unit[address].info &= 0xF8;//update info from offset 2 to 3
            }
            // prepare write data ,byte write to 4 bytes write
            UART_PRINTF("wr address : %d,offset : %d,data : %x\r\n", address, flash_unit_info_transf_data_idx(ret_info), data);
            w_tmp_buffer[0] = ret_info; // Get_page_info_in_SRAM(area,address);
            w_tmp_buffer[flash_unit_info_transf_data_idx(ret_info)] = data;
            flash_write(FLASH_SPACE_TYPE_NVR, area + address, 4, &w_tmp_buffer[0]);
            // UART_PRINTF("rd addr : %d\r\n",address);
            // flash_read(FLASH_SPACE_TYPE_NVR,area+address,4,&r_tmp_buffer[0]);
            // UART_PRINTF("data : %x%x%x%x\r\n",r_tmp_buffer[0],r_tmp_buffer[1],r_tmp_buffer[2],r_tmp_buffer[3]);
        }
        // flash_read(FLASH_SPACE_TYPE_NVR, extMCU_USED_FLASH_AREA+address, 4, &tmp_buffer[0]);
        // return tmp_buffer[data_offset];
    }
    else // Flash no have data
    {
        update_page_info_data(area, address, 0xFE, data);
        w_tmp_buffer[0] = 0xFE; // empty -> offset 1
        w_tmp_buffer[1] = data;
        w_tmp_buffer[2] = 0xFF;
        w_tmp_buffer[3] = 0xFF;
        UART_PRINTF("wr address : %d,offset : %d,data : %x\r\n", address, flash_unit_info_transf_data_idx(0xfe), data);
        flash_write(FLASH_SPACE_TYPE_NVR, area + address, 4, &w_tmp_buffer[0]);
        // UART_PRINTF("rd addr : %d\r\n",address);
        // flash_read(FLASH_SPACE_TYPE_NVR,area+address,4,&r_tmp_buffer[0]);
        // UART_PRINTF("data : %x%x%x%x\r\n",r_tmp_buffer[0],r_tmp_buffer[1],r_tmp_buffer[2],r_tmp_buffer[3]);
    }
}
#if 1
/****************************************/
/* internal parameter load/save utility */
/****************************************/
/*********************************************/
/*  input device_tag search last page no     */
/*********************************************/
uint8_t Get_last_page_no(uint8_t device_tag)
{
    uint8_t page_no = 0, data;
    uint8_t last_page_no = 0xFF;

    for (page_no = 0; page_no < 128;)
    {
        data = RFLASH_process(page_no, BK3432_USED_FLASH_AREA);
        if (data == 0xFF) // data is empty
        {
            return last_page_no;
        }
        else
        {
            //#define TAG_ADV                0xAA
            //#define TAG_BAUD               0xAB
            //#define TAG_NAME               0xAC
            //#define TAG_FFF0               0xB0
            //#define TAG_FFF1               0xB1
            //#define TAG_FFF2               0xB2

            if (data != device_tag)
            {
                if ((data < TAG_MAX) && (data >= TAG_ADV))
                {
                    page_no++;
                    data = RFLASH_process(page_no, BK3432_USED_FLASH_AREA); // get len
                    page_no += (data + 1);
                }
                else
                {
                    // error case
                }
            }
            else
            {
                last_page_no = page_no;
                page_no++;
                data = RFLASH_process(page_no, BK3432_USED_FLASH_AREA); // get len
                page_no += (data + 1);
            }
        }
    }
    return last_page_no;
}
/**********************************************************/
/*  input device_tag get this data of fallow this tag     */
/*  ex : TAG_ADV len = 1  data = 2,return data            */
/**********************************************************/
uint8_t Get_Tag_data_one_byte(uint8_t TAG_page_no)
{
    return (RFLASH_process(TAG_page_no + 2, BK3432_USED_FLASH_AREA)); // get data
}
/**********************************************************/
/*  input device_tag get this data of fallow this tag     */
/*  ex : TAG_ADV len = 2  data = 2,3,return data            */
/**********************************************************/
uint16_t Get_Tag_data_two_byte(uint8_t TAG_page_no)
{
    uint16_t ret_value = 0;
    ret_value = RFLASH_process(TAG_page_no + 2, BK3432_USED_FLASH_AREA);
    ret_value <<= 8;
    ret_value |= RFLASH_process(TAG_page_no + 3, BK3432_USED_FLASH_AREA);
    return ret_value;
}
/**********************************************************/
/*  input device_tag get this data of fallow this tag     */
/*  ex : TAG_ADV len = 20  data = "Serial-SPP" return data            */
/**********************************************************/
uint8_t Get_Tag_data_multi_byte(uint8_t TAG_page_no, uint8_t *data, uint8_t max_size)
{
    uint8_t page_no, buf_idx = 0;
    // TAG_len = RFLASH_process(TAG_page_no+1,BK3432_USED_FLASH_AREA);
    page_no = TAG_page_no + 2;

    for (buf_idx = 0; buf_idx < 20; buf_idx++)
    {
        data[buf_idx] = RFLASH_process(page_no + buf_idx, BK3432_USED_FLASH_AREA);
        // UART_PRINTF("page_no+buf_idx : %d,%c\r\n",page_no+buf_idx,data[buf_idx]);
    }
    return buf_idx;
}
/***************************************************************/
/* read page info from flash and return data index             */
/**************************************************************/
uint8_t Get_page_info_data_in_FLASH(uint16_t area, uint8_t page_no)
{
    uint8_t r_tmp_buffer[4];
    flash_read(FLASH_SPACE_TYPE_NVR, area + page_no, 4, &r_tmp_buffer[0]);
    // UART_PRINTF("%s,area : %x,page_no : %d,data:%x\r\n",__func__,area,page_no,r_tmp_buffer[0]);

    return r_tmp_buffer[0];
}

void Write_one_byte_to_flash(uint8_t TAG_value, uint8_t data, uint16_t area)
{
    uint8_t page_no, wdata;

    UART_PRINTF("%s\r\n", __func__);
    page_no = search_TAG_empty_page_no(TAG_value);
    UART_PRINTF("TAG_value : %x,w_page_no : %d,data : %x\r\n", TAG_value, page_no, data);
    wdata = TAG_value; // TAG
    WFLASH_process(page_no, wdata, area);
    page_no++;
    wdata = 1; // LEN
    WFLASH_process(page_no, wdata, area);
    page_no++;
    wdata = data;
    WFLASH_process(page_no, wdata, area);
}

void Write_two_bytes_to_flash(uint8_t TAG_value, uint16_t data, uint16_t area)
{
    uint8_t page_no, wdata;
    UART_PRINTF("%s\r\n", __func__);
    page_no = search_TAG_empty_page_no(TAG_value);
    UART_PRINTF("TAG_value : %x,w_page_no : %d,data : %x\r\n", TAG_value, page_no, data);
    wdata = TAG_value; // TAG
    WFLASH_process(page_no, wdata, area);
    page_no++;
    wdata = 2; // LEN
    WFLASH_process(page_no, wdata, area);
    page_no++;
    wdata = data >> 8; // value high
    WFLASH_process(page_no, wdata, area);
    page_no++;
    wdata = data; // value low
    WFLASH_process(page_no, wdata, area);
}

void Write_TAG_NAME_to_flash(uint8_t *data, uint16_t area)
{
    uint8_t page_no, wdata, i;
    page_no = search_TAG_empty_page_no(TAG_NAME);
    UART_PRINTF("%s,w_page_no : %d\r\n", __func__, page_no);
    wdata = TAG_NAME; // TAG
    WFLASH_process(page_no, wdata, area);
    page_no++;
    wdata = 20; // LEN
    WFLASH_process(page_no, wdata, area);
    page_no++;
    for (i = 0; i < 20; i++)
    {
        wdata = data[i];
        WFLASH_process(page_no + i, wdata, area);
    }
}
uint8_t search_TAG_empty_page_no(uint8_t TAG_value) // empty_page_no is 128 mean no free space
{
    uint8_t working_page_no, data, len, info;
    uint8_t empty_page_no = 128;
    uint8_t ret_info;
    uint8_t ret_data;
    // uint8_t result = 0;
    UART_PRINTF("%s\r\n", __func__);
    for (working_page_no = 0; working_page_no < 128;)
    {
        data = RFLASH_process(working_page_no, BK3432_USED_FLASH_AREA);
        UART_PRINTF("Read_data : %x\r\n", data);
        if (data == TAG_value)
        {
            // info = Get_page_info_data_in_FLASH(working_page_no,BK3432_USED_FLASH_AREA);
            Get_page_info_data(BK3432_USED_FLASH_AREA, working_page_no, &ret_data, &ret_info);
            UART_PRINTF("pageinfo : %x,TAG_value : %x\r\n", info, TAG_value);
            if (ret_info < 3)
            {
                empty_page_no = working_page_no;
                break;
            }
            else // this tag is full
            {
                UART_PRINTF("info is full\r\n");
                len = RFLASH_process(working_page_no + 1, BK3432_USED_FLASH_AREA);
                working_page_no += (len + 2); // change to next tag start
            }
        }
        else if ((data < TAG_MAX) && (data >= TAG_ADV))
        {

            len = RFLASH_process(working_page_no + 1, BK3432_USED_FLASH_AREA);
            working_page_no += (len + 2); // change to next tag start
        }
        if (data == 0xff) // empty data
        {
            empty_page_no = working_page_no;
            // UART_PRINTF("empty page no : %d\r\n",empty_page_no);
            break;
        }
        else
        {
            // error case not TAG
        }
    }
    return empty_page_no;
}

void save_parameter_to_BK3432_USED_FLASH_AREA(uint8_t TAG_value, uint8_t *buffer)
{
    uint8_t w_page_no = 0, last_page_no = 0, TAG_browse = 0;
    uint8_t backup_TAG_ADV, backup_TAG_BAUD;
    uint8_t *backup_TAG_NAME = NULL;
    uint16_t backup_TAG_FFF0 = 0, backup_TAG_FFF1 = 0, backup_TAG_FFF2 = 0;

    UART_PRINTF("TAG_value : %x\r\n", TAG_value);
    if ((TAG_value < TAG_MAX) && (TAG_value >= TAG_ADV))
    {
        w_page_no = search_TAG_empty_page_no(TAG_value);
        UART_PRINTF("empty_page_no : %d\r\n", w_page_no);
        if (w_page_no == 128) // no free space
        {
        restore_data:
            UART_PRINTF("Start restore\r\n");
            for (TAG_browse = TAG_ADV; TAG_browse < TAG_MAX; TAG_browse++)
            {
                last_page_no = Get_last_page_no(TAG_browse); // 1 step search
                if (last_page_no != 0xFF)                    // this tag found
                {
                    switch (TAG_browse)
                    {
                    case TAG_ADV:
                        backup_TAG_ADV = Get_Tag_data_one_byte(last_page_no);
                        UART_PRINTF("backup_adv : %x\r\n", backup_TAG_ADV);
                        if (TAG_value == TAG_ADV)
                        {
                            backup_TAG_ADV = buffer[0];
                        }
                        break;
                    case TAG_BAUD:
                        backup_TAG_BAUD = Get_Tag_data_one_byte(last_page_no);
                        UART_PRINTF("backup_TAG_BAUD : %x\r\n", backup_TAG_BAUD);
                        if (TAG_value == TAG_BAUD)
                        {
                            backup_TAG_BAUD = buffer[0];
                        }
                        break;
                    case TAG_FFF0:
                        backup_TAG_FFF0 = Get_Tag_data_two_byte(last_page_no);
                        if (TAG_value == TAG_FFF0)
                        {
                            backup_TAG_FFF0 = buffer[0];
                            backup_TAG_FFF0 <<= 8;
                            backup_TAG_FFF0 |= buffer[1];
                        }
                        break;
                    case TAG_FFF1:
                        backup_TAG_FFF1 = Get_Tag_data_two_byte(last_page_no);
                        if (TAG_value == TAG_FFF1)
                        {
                            backup_TAG_FFF1 = buffer[0];
                            backup_TAG_FFF1 <<= 8;
                            backup_TAG_FFF1 |= buffer[1];
                        }
                        break;
                    case TAG_FFF2:
                        backup_TAG_FFF2 = Get_Tag_data_two_byte(last_page_no);
                        if (TAG_value == TAG_FFF2)
                        {
                            backup_TAG_FFF2 = buffer[0];
                            backup_TAG_FFF2 <<= 8;
                            backup_TAG_FFF2 |= buffer[1];
                        }
                        break;
                    case TAG_NAME:
                        backup_TAG_NAME = ke_malloc(20, KE_MEM_NON_RETENTION);
                        Get_Tag_data_multi_byte(last_page_no, (uint8_t *)backup_TAG_NAME, APP_DEVICE_NAME_LENGTH_MAX);
                        if (TAG_value == TAG_NAME)
                        {
                            memcpy(backup_TAG_NAME, buffer, APP_DEVICE_NAME_LENGTH_MAX);
                        }
                        break;
                    }
                }
            }
            flash_erase(FLASH_SPACE_TYPE_NVR, BK3432_USED_FLASH_AREA, 0x200); // erase NVR 0x8000
            if (backup_TAG_ADV != 0)
            {
                Write_one_byte_to_flash(TAG_ADV, backup_TAG_ADV, BK3432_USED_FLASH_AREA);
            }
            if (backup_TAG_BAUD != 0)
            {
                Write_one_byte_to_flash(TAG_BAUD, backup_TAG_BAUD, BK3432_USED_FLASH_AREA);
            }
            if (backup_TAG_FFF0 != 0)
            {
                Write_two_bytes_to_flash(TAG_FFF0, backup_TAG_FFF0, BK3432_USED_FLASH_AREA);
            }
            if (backup_TAG_FFF1 != 0)
            {
                Write_two_bytes_to_flash(TAG_FFF1, backup_TAG_FFF1, BK3432_USED_FLASH_AREA);
            }
            if (backup_TAG_FFF2 != 0)
            {
                Write_two_bytes_to_flash(TAG_FFF2, backup_TAG_FFF2, BK3432_USED_FLASH_AREA);
            }
            if (backup_TAG_NAME != NULL)
            {
                Write_TAG_NAME_to_flash(backup_TAG_NAME, BK3432_USED_FLASH_AREA);
                ke_free(backup_TAG_NAME);
            }
        }
        else
        {
            uint8_t no_space = 0;

            // boundary check
            switch (TAG_value)
            {
            case TAG_ADV:
            case TAG_BAUD:
                if ((w_page_no + 1 + 1) > 127)
                {
                    no_space = 1; // no space to write
                }
                else
                    Write_one_byte_to_flash(TAG_value, buffer[0], BK3432_USED_FLASH_AREA);
                break;
            case TAG_NAME:
                if ((w_page_no + 1 + 20) > 127)
                {
                    no_space = 1; // no space to write
                }
                else
                {
                    Write_TAG_NAME_to_flash(buffer, BK3432_USED_FLASH_AREA);
                }
                break;
            case TAG_FFF0:
            case TAG_FFF1:
            case TAG_FFF2:
            {
                uint16_t w_data;
                w_data = buffer[0];
                w_data <<= 8;
                w_data |= buffer[1];
                if ((w_page_no + 1 + 2) > 127)
                {
                    no_space = 1; // no space to write
                }
                else
                    Write_two_bytes_to_flash(TAG_value, w_data, BK3432_USED_FLASH_AREA);
            }
            break;
            }
            if (no_space)
            {
                UART_PRINTF("no_space=1\r\n");
                goto restore_data;
            }
        }
    }
    else
    {
        UART_PRINTF("TAG value not found\r\n"); // error case
    }
}
void load_parameter_from_BK3432_USED_FLASH_AREA(void)
{
    uint8_t TAG_page_no, TAG_data;
    /*
        uint16_t Sram_Adv_interval = 160;
        uint16_t Sram_FFF0_Service = 0xFFF0;
        uint16_t Sram_FFF1_Char_Service = 0xFFF1;
        uint16_t Sram_FFF2_Char_Service = 0xFFF2;
        uint8_t  Sram_Baudrate;
    */
    // load TAG_ADV parameter
    UART_PRINTF("!!!%s!!!\r\n", __func__);
    TAG_page_no = Get_last_page_no(TAG_ADV);
    // if TAG_page_no == 0 TAG_ADV not found,used power on initial value
    if (TAG_page_no != 0xFF)
    {
        TAG_data = Get_Tag_data_one_byte(TAG_page_no);
        switch (TAG_data)
        {
        case 1:
            Sram_Adv_interval = 160; // 100ms
            break;
        case 2:
            Sram_Adv_interval = 400; // 250ms
            break;
        case 3:
            Sram_Adv_interval = 800; // 500ms
            break;
        case 4:
            Sram_Adv_interval = 2560; // 1600ms
            break;
        case 5:
            Sram_Adv_interval = 5120; // 3200ms
            break;
        }
        UART_PRINTF("Sram_Adv_interval : %d\r\n", Sram_Adv_interval);
    }
    // load TAG_BAUD parameter
    TAG_page_no = Get_last_page_no(TAG_BAUD);
    UART_PRINTF("Search TAG_BAUD,TAG_page_no : %x\r\n", TAG_page_no);
    // if TAG_page_no == 0 TAG_BAUD not found,used power on initial value
    if (TAG_page_no != 0xFF)
    {
        TAG_data = Get_Tag_data_one_byte(TAG_page_no);
        Sram_Baudrate = TAG_data;
    }
    UART_PRINTF("Sram_Baudrate : %d\r\n", Sram_Baudrate);
    // load TAG_NAME parameter
    TAG_page_no = Get_last_page_no(TAG_NAME);
    UART_PRINTF("Search TAG_NAME,TAG_page_no : %x\r\n", TAG_page_no);
    // if TAG_page_no == 0 TAG_BAUD not found,used power on initial value
    if (TAG_page_no != 0xFF)
    {
        memset(Sram_Device_name, 0, APP_DEVICE_NAME_LENGTH_MAX); // clear Sram_Device_name
        Get_Tag_data_multi_byte(TAG_page_no, (uint8_t *)Sram_Device_name, APP_DEVICE_NAME_LENGTH_MAX);
        UART_PRINTF("Device name : %s\r\n", Sram_Device_name);
    }
    // load TAG_FFF0 parameter
    TAG_page_no = Get_last_page_no(TAG_FFF0);
    if (TAG_page_no != 0xFF)
    {
        if (Sram_FFFx_load_from_flash & FFF0_LOAD_MASK)
        {
            Sram_FFF0_Service = Get_Tag_data_two_byte(TAG_page_no);
            UART_PRINTF("FFF0_para :%x\r\n", Sram_FFF0_Service);
        }
    }
    // load TAG_FFF1 parameter
    TAG_page_no = Get_last_page_no(TAG_FFF1);
    if (TAG_page_no != 0xFF)
    {
        if (Sram_FFFx_load_from_flash & FFF1_LOAD_MASK)
        {
            Sram_FFF1_Char_Service = Get_Tag_data_two_byte(TAG_page_no);
            UART_PRINTF("FFF1_para :%x\r\n", Sram_FFF1_Char_Service);
        }
    }
    // load TAG_FFF2 parameter
    TAG_page_no = Get_last_page_no(TAG_FFF2);
    if (TAG_page_no != 0xFF)
    {
        if (Sram_FFFx_load_from_flash & FFF2_LOAD_MASK)
        {
            Sram_FFF2_Char_Service = Get_Tag_data_two_byte(TAG_page_no);
            UART_PRINTF("FFF2_para :%x\r\n", Sram_FFF2_Char_Service);
        }
    }
}

void at_cmd_rsp(void)
{
    uint8_t len;
    extern uint8_t rxdata_buffer_len;

    if (uart_rx_done == 1)
    {
        len = rxdata_buffer_len;
        rxdata_buffer_len = 0;
        if (rxdata_buffer[len - 1] == '\n')
        {
            // AT command finish
            memcpy(&AT_cmd_buf[recv_AT_cmd_idx], rxdata_buffer, len);
            AT_cmd_len += len;
            recv_AT_cmd_idx = 0;
        }
        else
        {
            // command not finish
            memcpy(&AT_cmd_buf[recv_AT_cmd_idx], rxdata_buffer, len);
            recv_AT_cmd_idx = len;
            AT_cmd_len += len;
            uart_rx_done = 0;
            return;
        }
    }
}

at_rsp_content *at_wait_for_rsp(void)
{
    cpu_delay(15);
    while (true)
    {
        rwip_schedule();
        if (rxdata_buffer_len != 0)
        {
            rsp_result.len = rxdata_buffer_len;
            rsp_result.data = rxdata_buffer;
            break;
        }
    }

    return (&rsp_result);
}
#endif

#endif //(BLE_APP_PRESENT)

/// @} APP
