/**
 ****************************************************************************************
 *
 * @file app_midi.c
 *
 * @brief MIDI Application Module entry point
 *
 * @auth  gang.cheng
 *
 * @date  2016.05.31
 *
 * Copyright (C) Beken 2009-2016
 *
 *
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @addtogroup APP
 * @{
 ****************************************************************************************
 */

#include "rwip_config.h" // SW configuration

/*
 * INCLUDE FILES
 ****************************************************************************************
 */
#include <string.h>
#include "app.h"      // Application Definitions
#include "app_task.h" // application task definitions
#include "app_midi.h"
#include "midi_task.h" // health thermometer functions
#include "co_bt.h"
#include "prf_types.h" // Profile common types definition
#include "arch.h"      // Platform Definitions
#include "prf.h"
#include "midi.h"
#include "ke_timer.h"
#include "uart.h"
#include "uart2.h"
#include "gpio.h"

/*
 * DEFINES
 ****************************************************************************************
 */

/*
 * GLOBAL VARIABLE DEFINITIONS
 ****************************************************************************************
 */

/// f000 Application Module Environment Structure
struct app_midi_env_tag app_midi_env;

/*
 * GLOBAL FUNCTION DEFINITIONS
 ****************************************************************************************
 */

void app_midi_init(void)
{

	// Reset the environment
	memset(&app_midi_env, 0, sizeof(struct app_midi_env_tag));
}

void app_midi_add_midis(void)
{

	struct midi_db_cfg *db_cfg;

	struct gapm_profile_task_add_cmd *req = KE_MSG_ALLOC_DYN(GAPM_PROFILE_TASK_ADD_CMD,
			 TASK_GAPM, TASK_APP,
			 gapm_profile_task_add_cmd, sizeof(struct midi_db_cfg));
	// Fill message
	req->operation = GAPM_PROFILE_TASK_ADD;
	req->sec_lvl = 0;
	req->prf_task_id = TASK_ID_MIDI;
	req->app_task = TASK_APP;
	req->start_hdl = 0; // req->start_hdl = 0; dynamically allocated

	// Set parameters
	db_cfg = (struct midi_db_cfg *)req->param;

	// Sending of notifications is supported
	db_cfg->features = MIDI_NTF_SUP;

	// Send the message
	ke_msg_send(req);
}

void app_midi_send_payload(uint8_t *buf, uint8_t len)
{
	// Allocate the message
	struct midi_payload_upd_req *req = KE_MSG_ALLOC(MIDI_PAYLOAD_UPD_REQ,
		                       prf_get_task_from_id(TASK_ID_MIDI),
		                       TASK_APP,
		                       midi_payload_upd_req);
	// Fill in the parameter structure
	req->length = len;
	memcpy(req->tx_payload, buf, len);

	// Send the message
	ke_msg_send(req);
}

static int midi_payload_ntf_cfg_ind_handler(ke_msg_id_t const msgid,
	                   struct midi_payload_ntf_cfg_ind const *param,
	                   ke_task_id_t const dest_id,
	                   ke_task_id_t const src_id)
{
	UART_PRINTF("param->ntf_cfg = %x\r\n", param->ntf_cfg);
	if (param->ntf_cfg == PRF_CLI_STOP_NTFIND)
	{
		ke_timer_clear(MIDI_PAYLOAD_PERIOD_NTF, dest_id);
	}
	else
	{
		ke_timer_set(MIDI_PAYLOAD_PERIOD_NTF, dest_id, 1);
	}

	return (KE_MSG_CONSUMED);
}

static int midi_payload_upd_handler(ke_msg_id_t const msgid,
	           struct midi_payload_upd_rsp const *param,
	           ke_task_id_t const dest_id,
	           ke_task_id_t const src_id)
{
	return (KE_MSG_CONSUMED);
}

/**
 ****************************************************************************************
 * @brief
 *
 * @param[in] msgid     Id of the message received.
 * @param[in] param     Pointer to the parameters of the message.
 * @param[in] dest_id   ID of the receiving task instance (TASK_GAP).
 * @param[in] src_id    ID of the sending task instance.
 *
 * @return If the message was consumed or not.
 ****************************************************************************************
 */
static int app_midi_msg_dflt_handler(ke_msg_id_t const msgid,
	         void const *param,
	         ke_task_id_t const dest_id,
	         ke_task_id_t const src_id)
{
	// Drop the message
	return (KE_MSG_CONSUMED);
}

static int midi_writer_req_handler(ke_msg_id_t const msgid,
	          struct midi_writer_ind *param,
	          ke_task_id_t const dest_id,
	          ke_task_id_t const src_id)
{
	// bypass payload to MCU
  uint8_t recv_data[256] = {0};
  memcpy(recv_data, param->rx_payload, param->length);
  recv_data[param->length] = 0x0D;    //CR
  recv_data[param->length+1] = 0x0A;  //LF
	//uart_send((unsigned char *)param->rx_payload, param->length);
  uart_send((unsigned char *)recv_data, param->length + 2);

	return (KE_MSG_CONSUMED);
}

static int midi_period_ntf_handler(ke_msg_id_t const msgid,
	          struct midi_payload_ntf_cfg_ind const *param,
	          ke_task_id_t const dest_id,
	          ke_task_id_t const src_id)
{
  return (KE_MSG_CONSUMED);
}

/*
 * LOCAL VARIABLE DEFINITIONS
 ****************************************************************************************
 */

/// Default State handlers definition
const struct ke_msg_handler app_midi_msg_handler_list[] =
    {
        // Note: first message is latest message checked by kernel so default is put on top.
        {KE_MSG_DEFAULT_HANDLER, (ke_msg_func_t)app_midi_msg_dflt_handler},
        {MIDI_PAYLOAD_NTF_CFG_IND, (ke_msg_func_t)midi_payload_ntf_cfg_ind_handler},
        {MIDI_PAYLOAD_UPD_RSP, (ke_msg_func_t)midi_payload_upd_handler},
        {MIDI_WRITER_REQ_IND, (ke_msg_func_t)midi_writer_req_handler},
        {MIDI_PAYLOAD_PERIOD_NTF, (ke_msg_func_t)midi_period_ntf_handler},
};

const struct ke_state_handler app_midi_table_handler =
    {&app_midi_msg_handler_list[0], (sizeof(app_midi_msg_handler_list) / sizeof(struct ke_msg_handler))};
