/**
 ****************************************************************************************
 *
 * @file custom.c
 *
 * @brief Device Information Service Server Implementation.
 *
 * Copyright (C) RivieraWaves 2009-2016
 *
 *
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @addtogroup CUSTOM
 * @{
 ****************************************************************************************
 */

/*
 * INCLUDE FILES
 ****************************************************************************************
 */
#include "rwip_config.h"

#if (BLE_CUSTOM_SERVER)
#include "attm.h"
#include "custom.h"
#include "custom_task.h"
#include "prf_utils.h"
#include "prf.h"

#include "ke_mem.h"

/*
 * MACROS
 ****************************************************************************************
 */

/// Maximal length for Characteristic values - 128 bytes
#define CUSTOM_VAL_MAX_LEN                         (128)

/*
 * CUSTOM ATTRIBUTES
 ****************************************************************************************
 */

/// Full CUSTOM Database Description - Used to add attributes into the database
const struct attm_desc custom_att_db[CUSTOM_IDX_NB] =
{
    // Device Information Service Declaration
    [CUSTOM_IDX_SVC]                           =   {ATT_DECL_PRIMARY_SERVICE, PERM(RD, ENABLE), 0, 0},

    // Manufacturer Name Characteristic Declaration
    [CUSTOM_IDX_REMOTE_IO_CHAR]                =   {ATT_DECL_CHARACTERISTIC, PERM(RD, ENABLE), 0, 0},
    // Manufacturer Name Characteristic Value
    [CUSTOM_IDX_REMOTE_IO_VAL]                 =   {ATT_USER_SERVER_CHAR_TEST1, PERM(RD, ENABLE) | PERM(WRITE_COMMAND, ENABLE), \
                                                    PERM(RI, ENABLE), CUSTOM_VAL_MAX_LEN},
    
};



/*
 * LOCAL FUNCTION DEFINITIONS
 ****************************************************************************************
 */

/**
 ****************************************************************************************
 * @brief Initialization of the CUSTOM module.
 * This function performs all the initializations of the Profile module.
 *  - Creation of database (if it's a service)
 *  - Allocation of profile required memory
 *  - Initialization of task descriptor to register application
 *      - Task State array
 *      - Number of tasks
 *      - Default task handler
 *
 * @param[out]    env        Collector or Service allocated environment data.
 * @param[in|out] start_hdl  Service start handle (0 - dynamically allocated), only applies for services.
 * @param[in]     app_task   Application task number.
 * @param[in]     sec_lvl    Security level (AUTH, EKS and MI field of @see enum attm_value_perm_mask)
 * @param[in]     param      Configuration parameters of profile collector or service (32 bits aligned)
 *
 * @return status code to know if profile initialization succeed or not.
 ****************************************************************************************
 */
static uint8_t custom_init (struct prf_task_env* env, uint16_t* start_hdl, uint16_t app_task, uint8_t sec_lvl,  struct custom_db_cfg* params)
{
    //------------------ create the attribute database for the profile -------------------
  
    // Service content flag
    uint32_t cfg_flag;
    // DB Creation Statis
    uint8_t status = ATT_ERR_NO_ERROR;

    // Compute Attribute Table and save it in environment
    cfg_flag = custom_compute_cfg_flag(params->features);

    status = attm_svc_create_db(start_hdl, ATT_USER_SERVER_CUSTOM, (uint8_t *)&cfg_flag,
            CUSTOM_IDX_NB, NULL, env->task, &custom_att_db[0],
            (sec_lvl & (PERM_MASK_SVC_DIS | PERM_MASK_SVC_AUTH | PERM_MASK_SVC_EKS)));

    //-------------------- allocate memory required for the profile  ---------------------
    if (status == ATT_ERR_NO_ERROR)
    {
        struct custom_env_tag* custom_env =
                (struct custom_env_tag* ) ke_malloc(sizeof(struct custom_env_tag), KE_MEM_ATT_DB);

        // Remote I/O Level characteristic value permissions
        //uint16_t perm = PERM(RD, ENABLE) | PERM(WRITE_REQ, ENABLE);
        //attm_att_set_permission(*start_hdl + CUSTOM_IDX_REMOTE_IO_VAL, perm, 0);
        
        // allocate CUSTOM required environment variable
        env->env = (prf_env_t*) custom_env;
        custom_env->start_hdl = *start_hdl;
        custom_env->features  = params->features;
        custom_env->prf_env.app_task = app_task
                | (PERM_GET(sec_lvl, SVC_MI) ? PERM(PRF_MI, ENABLE) : PERM(PRF_MI, DISABLE));
        custom_env->prf_env.prf_task = env->task | PERM(PRF_MI, DISABLE);

        // initialize environment variable
        env->id                     = TASK_ID_CUSTOM;
        env->desc.idx_max           = CUSTOM_IDX_MAX;
        env->desc.state             = custom_env->state;
        env->desc.default_handler   = &custom_default_handler;
        co_list_init(&(custom_env->values));

        // service is ready, go into an Idle state
        ke_state_set(env->task, CUSTOM_IDLE);
    }

    return (status);
}
/**
 ****************************************************************************************
 * @brief Destruction of the CUSTOM module - due to a reset for instance.
 * This function clean-up allocated memory (attribute database is destroyed by another
 * procedure)
 *
 * @param[in|out]    env        Collector or Service allocated environment data.
 ****************************************************************************************
 */
static void custom_destroy(struct prf_task_env* env)
{
    struct custom_env_tag* custom_env = (struct custom_env_tag*) env->env;

    // remove all values present in list
    while(!co_list_is_empty(&(custom_env->values)))
    {
        struct co_list_hdr* hdr = co_list_pop_front(&(custom_env->values));
        ke_free(hdr);
    }

    // free profile environment variables
    env->env = NULL;
    ke_free(custom_env);
}

/**
 ****************************************************************************************
 * @brief Handles Connection creation
 *
 * @param[in|out]    env        Collector or Service allocated environment data.
 * @param[in]        conidx     Connection index
 ****************************************************************************************
 */
static void custom_create(struct prf_task_env* env, uint8_t conidx)
{
    /* Nothing to do */
}

/**
 ****************************************************************************************
 * @brief Handles Disconnection
 *
 * @param[in|out]    env        Collector or Service allocated environment data.
 * @param[in]        conidx     Connection index
 * @param[in]        reason     Detach reason
 ****************************************************************************************
 */
static void custom_cleanup(struct prf_task_env* env, uint8_t conidx, uint8_t reason)
{
    /* Nothing to do */
}

/*
 * GLOBAL VARIABLE DEFINITIONS
 ****************************************************************************************
 */

/// CUSTOM Task interface required by profile manager
const struct prf_task_cbs custom_itf =
{
        (prf_init_fnct) custom_init,
        custom_destroy,
        custom_create,
        custom_cleanup,
};

/*
 * FUNCTION DEFINITIONS
 ****************************************************************************************
 */

const struct prf_task_cbs* customs_prf_itf_get(void)
{
   return &custom_itf;
}

uint32_t custom_compute_cfg_flag(uint16_t features)
{
    //Service Declaration
    uint32_t cfg_flag = 1;

    for (uint8_t i = 0; i<CUSTOM_CHAR_MAX; i++)
    {
        if (((features >> i) & 1) == 1)
        {
            cfg_flag |= (3 << (i*2 + 1));
        }
    }

    return cfg_flag;
}


uint8_t custom_handle_to_value(struct custom_env_tag* env, uint16_t handle)
{
    uint8_t value = CUSTOM_CHAR_MAX;

    // handle cursor, start from first characteristic of service handle
    uint16_t cur_hdl = env->start_hdl + 1;

    for (uint8_t i = 0; i<CUSTOM_CHAR_MAX; i++)
    {
        if (((env->features >> i) & 1) == 1)
        {
            // check if value handle correspond to requested handle
            if((cur_hdl +1) == handle)
            {
                value = i;
                break;
            }
            cur_hdl += 2;
        }
    }

    return value;
}

uint16_t custom_value_to_handle(struct custom_env_tag* env, uint8_t value)
{
    uint16_t handle = env->start_hdl + 1;
    int8_t i;

    for (i = 0; i<CUSTOM_CHAR_MAX; i++)
    {
        if (((env->features >> i) & 1) == 1)
        {
            // requested value
            if(value == i)
            {
                handle += 1;
                break;
            }
            handle += 2;
        }
    }

    // check if handle found
    return ((i == CUSTOM_CHAR_MAX) ? ATT_INVALID_HDL : handle);
}

uint8_t custom_check_val_len(uint8_t char_code, uint8_t val_len)
{
    uint8_t status = GAP_ERR_NO_ERROR;

    // Check if length is upper than the general maximal length
    if (val_len > CUSTOM_VAL_MAX_LEN)
    {
        status = PRF_ERR_UNEXPECTED_LEN;
    }    

    return (status);
}



#endif //BLE_CUSTOM_SERVER

/// @} CUSTOM
