/**
 ****************************************************************************************
 *
 * @file f000.c
 *
 * @brief f000 Server Implementation.
 *
 * Copyright (C) beken 2009-2015
 *
 *
 ****************************************************************************************
 */

#include "rwip_config.h"

#if (BLE_MIDI_SERVER)
#include "attm.h"
#include "midi.h"
#include "midi_task.h"
#include "prf_utils.h"
#include "prf.h"
#include "ke_mem.h"
#include "attm_util128.h"
#include "uart.h"
#include "uart2.h"

/*
 * UART transparent ATTRIBUTES DEFINITION
 ****************************************************************************************
 */

/// Full UART Tansparent Database Description - Used to add attributes into the database
const struct attm_desc midi_att_db[MIDI_DATA_IO_IDX_NB] =

    {
        // UART Transparent Service Declaration
        [MIDI_IDX_SVC] = {ATT_DECL_PRIMARY_SERVICE, PERM(RD, ENABLE), 0, 0},
        //  Characteristic Value
        //[UART_TP_RX_CHAR] = {ATT_DECL_CHARACTERISTIC, PERM(RD, ENABLE), 0, 0},
        // UART Transparent Rx Characteristic Value
        //[UART_TP_RX_PAYLOAD] = {ATT_UART_TP_SERVER_CHAR_RX, PERM(WRITE_COMMAND, ENABLE), PERM(UUID_LEN, UUID_128)|PERM(RI, ENABLE)|ATT_UUID_128_LEN, UART_RX_PAYLOAD_LEN * sizeof(uint8_t)},

        //  Characteristic Value
        [MIDI_DATA_IO_CHAR] = {ATT_DECL_CHARACTERISTIC, PERM(RD, ENABLE), 0, 0},
        // UART Transparent Tx Characteristic Value
        [MIDI_DATA_IO_PAYLOAD] = {ATT_MIDI_SERVER_CHAR, PERM(RD, ENABLE) | PERM(WRITE_COMMAND, ENABLE) | PERM(NTF, ENABLE), PERM(UUID_LEN, UUID_128)|PERM(RI, ENABLE)|ATT_UUID_128_LEN, UART_TX_PAYLOAD_LEN * sizeof(uint8_t)},
        // UART Transparent Characteristic - Client Characteristic Configuration Descriptor
        [MIDI_DATA_IO_NTF_CFG] = {ATT_DESC_CLIENT_CHAR_CFG, PERM(RD, ENABLE) | PERM(WRITE_REQ, ENABLE), 0, 0},

}; /// Macro used to retrieve permission value from access and rights on attribute.

static uint8_t midi_init(struct prf_task_env *env, uint16_t *start_hdl, uint16_t app_task, uint8_t sec_lvl, struct midi_db_cfg *params)
{
    uint16_t shdl;
    struct midi_env_tag *midi_env = NULL;
    // Status
    uint8_t status = GAP_ERR_NO_ERROR;

    //-------------------- allocate memory required for the profile  ---------------------
    midi_env = (struct midi_env_tag *)ke_malloc(sizeof(struct midi_env_tag), KE_MEM_ATT_DB);
    memset(midi_env, 0, sizeof(struct midi_env_tag));

    // Service content flag
    uint8_t cfg_flag = MIDI_CFG_FLAG_MANDATORY_MASK;

    // Save database configuration
    midi_env->features |= (params->features);
    UART_PRINTF("%s,midi_env->features : %x\r\n", __func__, midi_env->features);
    // Check if notifications are supported
    if (params->features == MIDI_NTF_SUP)
    {
        cfg_flag |= MIDI_CFG_FLAG_NTF_SUP_MASK;
    }
    shdl = *start_hdl;

    // Create UART transparent in the DB
    //------------------ create the attribute database for the profile -------------------
    status = attm_util_svc_create_db128(&(shdl), ATT_MIDI_SERVER_SVC, (uint8_t *)&cfg_flag,
                                MIDI_DATA_IO_IDX_NB, NULL, env->task, &midi_att_db[0],
                                (sec_lvl & (PERM_MASK_SVC_DIS | PERM_MASK_SVC_AUTH | PERM_MASK_SVC_EKS)));

    //-------------------- Update profile task information  ---------------------
    if (status == ATT_ERR_NO_ERROR)
    {

        // allocate UART transparent required environment variable
        env->env = (prf_env_t *)midi_env;
        *start_hdl = shdl;
        midi_env->start_hdl = *start_hdl;
        midi_env->prf_env.app_task = app_task | (PERM_GET(sec_lvl, SVC_MI) ? PERM(PRF_MI, ENABLE) : PERM(PRF_MI, DISABLE));
        midi_env->prf_env.prf_task = env->task | PERM(PRF_MI, DISABLE);

        // initialize environment variable
        env->id = TASK_ID_MIDI;
        env->desc.idx_max = MIDI_IDX_MAX;
        env->desc.state = midi_env->state;
        env->desc.default_handler = &midi_default_handler;

        // service is ready, go into an Idle state
        ke_state_set(env->task, MIDI_IDLE);
    }
    else if (midi_env != NULL)
    {
        ke_free(midi_env);
    }

    return (status);
}

static void midi_destroy(struct prf_task_env *env)
{
    struct midi_env_tag *midi_env = (struct midi_env_tag *)env->env;

    // clear on-going operation
    if (midi_env->operation != NULL)
    {
        ke_free(midi_env->operation);
    }

    // free profile environment variables
    env->env = NULL;
    ke_free(midi_env);
}

static void midi_create(struct prf_task_env *env, uint8_t conidx)
{
    struct midi_env_tag *midi_env = (struct midi_env_tag *)env->env;
    ASSERT_ERR(conidx < BLE_CONNECTION_MAX);

    // force notification config to zero when peer device is connected
    midi_env->ntf_cfg[conidx] = 0;
}

static void midi_cleanup(struct prf_task_env *env, uint8_t conidx, uint8_t reason)
{
    struct midi_env_tag *midi_env = (struct midi_env_tag *)env->env;

    ASSERT_ERR(conidx < BLE_CONNECTION_MAX);
    // force notification config to zero when peer device is disconnected
    midi_env->ntf_cfg[conidx] = 0;
}

void midi_notify_payload(struct midi_env_tag *midi_env, struct midi_payload_upd_req const *param)
{
    // Allocate the GATT notification message
    struct gattc_send_evt_cmd *tx_payload = KE_MSG_ALLOC_DYN(GATTC_SEND_EVT_CMD,
                                                             KE_BUILD_ID(TASK_GATTC, 0), prf_src_task_get(&(midi_env->prf_env), 0),
                                                             gattc_send_evt_cmd, sizeof(uint8_t) * (param->length));
    // Fill in the parameter structure
    tx_payload->operation = GATTC_NOTIFY;
    tx_payload->handle = midi_get_att_handle(MIDI_DATA_IO_PAYLOAD);
    // pack measured value in database
    tx_payload->length = param->length;
    // tx_payload->value[0] = midi_env->tx_payload[0];
    memcpy(&tx_payload->value[0], &param->tx_payload[0], param->length);
    // send notification to peer device
    ke_msg_send(tx_payload);
}

/// BASS Task interface required by profile manager
const struct prf_task_cbs midi_itf =
    {
        (prf_init_fnct)midi_init,
        midi_destroy,
        midi_create,
        midi_cleanup,
};

const struct prf_task_cbs *midi_prf_itf_get(void)
{
    return &midi_itf;
}

uint16_t midi_get_att_handle(uint8_t att_idx)
{

    struct midi_env_tag *midi_env = PRF_ENV_GET(MIDI, midi);
    uint16_t handle = ATT_INVALID_HDL;

    handle = midi_env->start_hdl;

    // increment index according to expected index
    if (att_idx < MIDI_DATA_IO_NTF_CFG)
    {
        handle += att_idx;
    }
    // FFF1 notification
    else if ((att_idx == MIDI_DATA_IO_NTF_CFG) && (((midi_env->features) & 0x01) == MIDI_NTF_SUP))
    {
        handle += MIDI_DATA_IO_NTF_CFG;
    }
    else
    {
        handle = ATT_INVALID_HDL;
    }

    return handle;
}

uint8_t midi_get_att_idx(uint16_t handle, uint8_t *att_idx)
{
    struct midi_env_tag *midi_env = PRF_ENV_GET(MIDI, midi);
    uint16_t hdl_cursor = midi_env->start_hdl;
    uint8_t status = PRF_APP_ERROR;

    // Browse list of services
    // handle must be greater than current index
    // check if it's a mandatory index
    if (handle <= (hdl_cursor + MIDI_DATA_IO_PAYLOAD))
    {
        *att_idx = handle - hdl_cursor;
        status = GAP_ERR_NO_ERROR;
    }
    hdl_cursor += MIDI_DATA_IO_PAYLOAD;

    // check if it's a notify index
    UART_PRINTF("features : %x \r\n", midi_env->features);
    if (((midi_env->features) & 0x01) == MIDI_NTF_SUP)
    {
        hdl_cursor++;
        if (handle == hdl_cursor)
        {
            *att_idx = MIDI_DATA_IO_NTF_CFG;
            status = GAP_ERR_NO_ERROR;
        }
    }
    hdl_cursor++;

    return (status);
}

#endif // (BLE_MIDI_SERVER)
