/**
  **************************************************************************
  * @file     main.c
  * @version  v2.0.3
  * @date     2022-08-16
  * @brief    main program
  **************************************************************************
  *                       Copyright notice & Disclaimer
  *
  * The software Board Support Package (BSP) that is made available to
  * download from Artery official website is the copyrighted work of Artery.
  * Artery authorizes customers to use, copy, and distribute the BSP
  * software and its related documentation for the purpose of design and
  * development in conjunction with Artery microcontrollers. Use of the
  * software is governed by this copyright notice and the following disclaimer.
  *
  * THIS SOFTWARE IS PROVIDED ON "AS IS" BASIS WITHOUT WARRANTIES,
  * GUARANTEES OR REPRESENTATIONS OF ANY KIND. ARTERY EXPRESSLY DISCLAIMS,
  * TO THE FULLEST EXTENT PERMITTED BY LAW, ALL EXPRESS, IMPLIED OR
  * STATUTORY OR OTHER WARRANTIES, GUARANTEES OR REPRESENTATIONS,
  * INCLUDING BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
  * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.
  *
  **************************************************************************
  */
  
/* Includes ------------------------------------------------------------------*/
#include "at32wb415_board.h"
#include "at32wb415_clock.h"
#include "stdio.h"
#include "string.h"
#include "at_cmd.h"
#include "tp_mode.h"

//#define BT_FLASH_WR_TEST
#define SIZEOFMSG                        (sizeof(at_cmd_list)/sizeof(struct at_cmd_msg))
#define MIDI_STATUS_BYTE                 (2)
#define MIDI_NOTE_BYTE                   (3)
#define MIDI_VELOCITY_BYTE               (4)

extern cmd_flash_read_status fr_flag;
extern cmd_rsp_status cmd_result;
extern flag_status recv_cmp_flag;
extern uint8_t bt_flash_data;
extern char recv_data[];
extern uartstruct tp_mode_rx_uart;
extern uartstruct tp_mode_tx_uart;

flag_status UART_TP_MODE = RESET;
flag_status USB_TP_MODE = RESET;

/* maybe this flag is not necessary */
flag_status switch_mode_flag = RESET;

struct at_cmd_msg at_cmd_list[] =
{
  {AT_CMD_IO_SET_LOW, AT_CMD_IOSET0},
  {AT_CMD_IO_SET_HIGH, AT_CMD_IOSET1},
  {AT_CMD_IO_GET,  AT_CMD_IOGET},
};

static void uart_at_interface_init(uint32_t baudrate);
static void main_loop_func(void);
static void midiCommand(uint8_t status, uint8_t note, uint8_t velocity);
static uint8_t midiMessage[] = {0x80, 0x80, 0x00, 0x00, 0x00};
static uint8_t midiMelody[] = {64, 66, 71, 73, 74, 66, 64, 73, 71, 66, 74, 73};
const uint8_t bpm = 72;
float beatDuration = 60.0 / bpm * 1000;

/** @addtogroup AT32F415_StdPeriph_Examples
  * @{
  */

/** @addtogroup USART_Printf
  * @{
  */ 

/**
  * @brief  Main program
  * @param  None
  * @retval None
  */
int main(void)
{     
#ifdef BT_FLASH_WR_TEST
  uint8_t data;
#endif
  crm_clocks_freq_type crm_freq_struct = {0};
  system_clock_config();
  crm_clocks_freq_get(&crm_freq_struct);
  
  at32_board_init();
  at32_button_init();

  nvic_priority_group_config(NVIC_PRIORITY_GROUP_4);
  
  uart_at_interface_init(115200);
  
  uart_print_init(115200);
  printf("uart log ready...\r\n");
  
  /* Enable the USARTx Interrupt */
  usart_interrupt_enable(USART3, USART_RDBF_INT, TRUE);
  usart_enable(USART3, TRUE);
            
  while (1)
  {
    main_loop_func();
  }
}

/**
  * @brief  this function config at-command interface
  * @param  baudrate: uart baudrate
  * @retval none
  */
static void uart_at_interface_init(uint32_t baudrate)
{
  gpio_init_type gpio_init_struct;
 
  nvic_irq_enable(USART3_IRQn, 0, 0);
  
  crm_periph_clock_enable(CRM_USART3_PERIPH_CLOCK, TRUE);
  crm_periph_clock_enable(CRM_GPIOA_PERIPH_CLOCK, TRUE);  
  crm_periph_clock_enable(CRM_IOMUX_PERIPH_CLOCK, TRUE);
  
  gpio_pin_remap_config(USART3_GMUX_0010, TRUE);
  gpio_init_struct.gpio_drive_strength = GPIO_DRIVE_STRENGTH_STRONGER;
  gpio_init_struct.gpio_out_type  = GPIO_OUTPUT_PUSH_PULL;
  gpio_init_struct.gpio_mode = GPIO_MODE_MUX;
  gpio_init_struct.gpio_pins = GPIO_PINS_7;
  gpio_init_struct.gpio_pull = GPIO_PULL_NONE;
  gpio_init(GPIOA, &gpio_init_struct);
  
  gpio_init_struct.gpio_drive_strength = GPIO_DRIVE_STRENGTH_STRONGER;
  gpio_init_struct.gpio_out_type  = GPIO_OUTPUT_PUSH_PULL;
  gpio_init_struct.gpio_mode = GPIO_MODE_INPUT;
  gpio_init_struct.gpio_pins = GPIO_PINS_6;
  gpio_init_struct.gpio_pull = GPIO_PULL_UP;
  gpio_init(GPIOA, &gpio_init_struct);
  
  usart_init(USART3, baudrate, USART_DATA_8BITS, USART_STOP_1_BIT);
  usart_transmitter_enable(USART3, TRUE);
  usart_receiver_enable(USART3, TRUE);
  
  /* Enable the USARTx Interrupt */
  usart_interrupt_enable(USART3, USART_RDBF_INT, TRUE);
  usart_enable(USART3, TRUE); 
}

static void midiCommand(uint8_t status, uint8_t note, uint8_t velocity) 
{
  /* | header | timestamp | status | note | velocity | */
  // MIDI data goes in the last three bytes of the midiMessage array:
  midiMessage[MIDI_STATUS_BYTE] = status;
  midiMessage[MIDI_NOTE_BYTE] = note;
  midiMessage[MIDI_VELOCITY_BYTE] = velocity;
}

/**
  * @brief  this function handle general mode and transparent mode.
  * @param  none
  * @retval none
  */
static void main_loop_func(void)
{
  uint32_t i = 0, j = 0;
  uint32_t noteDuration = beatDuration / 4;
  if(UART_TP_MODE == RESET)
  {
    at_cmd_send(AT_CMD_TP_MODE_ON);
    while(cmd_rsp_get(AT_TP_MODE_OK1) != RSP_OK);
    UART_TP_MODE = SET;
    printf("enter bypass mode\r\n");
  }
  if(at32_button_press() == USER_BUTTON)
  {
    //printf("Button Pressed\r\n");
    for(j = 0; j < sizeof(midiMelody); j++)
    {
      midiCommand(0x90, midiMelody[j], 127);
      for(i = 0; i < sizeof(midiMessage); i++)
      {
        while(usart_flag_get(USART3, USART_TDBE_FLAG) !=SET);
        usart_data_transmit(USART3, midiMessage[i]);
      }
      delay_ms(noteDuration);
    }
  }
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t* file, uint32_t line)
{ 
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */

  /* Infinite loop */
  while (1)
  {}
}

#endif

/**
  * @}
  */ 

/**
  * @}
  */ 
